"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from re import compile

from encommon.types import DictStrAny
from encommon.utils import read_sample
from encommon.utils import read_text

from httpx import Response

from respx import MockRouter



SAMPLES = (
    Path(__file__).parent
    / 'samples')



def mock_hubi(
    respx_mock: MockRouter,
    replaces: DictStrAny,
) -> None:
    """
    Construct the mocker objects for simulating operations.

    :param respx_mock: Object for mocking request operation.
    :param replaces: Mapping of what to replace in samples.
    """

    samples = SAMPLES / 'origin'


    source = read_text(
        f'{samples}/source'
        '/jupiter.json')

    source = read_sample(
        sample=source,
        replace=replaces,
        prefix=False)

    (respx_mock
     .get(
         'https://192.168.1.11'
         '/apps/api/69'
         '/devices/all')
     .mock(Response(
         status_code=200,
         content=source)))


    source = read_text(
        f'{samples}/source'
        '/neptune.json')

    source = read_sample(
        sample=source,
        replace=replaces,
        prefix=False)

    (respx_mock
     .get(
         'https://192.168.2.11'
         '/apps/api/69'
         '/devices/all')
     .mock(Response(
         status_code=200,
         content=source)))


    match = compile(
        r'^https\S+\.168\.[12]\.11'
        r'\/apps\/api\/69\/devices'
        r'\/\d+\/(set|on|off)(.*|)$')

    (respx_mock
     .get(match)
     .mock(Response(200)))



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import save_text

from respx import MockRouter

from ....conftest import config_factory
from ....conftest import homie_factory

if TYPE_CHECKING:
    from ...homie import Homie
    from ....utils import TestTimes
    from ....utils import TestBodies



def test_HomieDesire(
    homie: 'Homie',
    bodies: 'TestBodies',
    times: 'TestTimes',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    :param times: Times that are used with tests in project.
    """

    childs = homie.childs
    desires = childs.desires


    planets = bodies.planets

    for planet in planets:

        name = f'{planet}_active'

        desire = desires[name]


        attrs = lattrs(desire)

        assert attrs == [
            '_HomieChild__homie',
            '_HomieChild__name',
            '_HomieChild__params',
            '_HomieDesire__wheres',
            '_HomieDesire__timers']


        assert inrepr(
            'desire.HomieDesire',
            desire)

        assert isinstance(
            hash(desire), int)

        assert instr(
            'desire.HomieDesire',
            desire)


        desire.validate()

        assert desire.homie

        assert desire.name == name

        assert desire.family == 'builtins'

        assert desire.kind == 'desire'

        assert desire.params

        assert desire.weight == 0

        assert len(desire.devices) == 1

        assert len(desire.groups) == 1

        assert len(desire.wheres) == 5

        assert desire.paused(
            desire.devices[0])

        assert desire.dumped



def test_HomieDesires_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
    times: 'TestTimes',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    :param times: Times that are used with tests in project.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        desires:
         pytest:
          store:
          - unique: desire
            value: true
            expire: '1y'
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    childs = homie.childs

    desire = (
        childs.desires
        ['pytest'])


    assert not desire.devices

    assert not desire.groups


    time = times.middle

    assert not (
        desire.when(time))



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import save_text

from respx import MockRouter

from ..change import DriverPhueChange
from ....conftest import config_factory
from ....conftest import homie_factory

if TYPE_CHECKING:
    from ....homie import Homie
    from ....utils import TestBodies
    from ....utils import TestTimes



def test_DriverPhueChange(
    homie: 'Homie',
    bodies: 'TestBodies',
    times: 'TestTimes',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    :param times: Times that are used with tests in project.
    """

    childs = homie.childs
    desires = childs.desires


    assert homie.refresh()


    planets = bodies.planets

    family = 'philips'

    for planet in planets:

        desire = desires[
            f'{planet}_active']

        driver = (
            desire.wheres[1]
            .drivers[0])

        assert isinstance(
            driver, DriverPhueChange)


        attrs = lattrs(driver)

        assert attrs == [
            '_HomieDriver__plugin',
            '_HomieDriver__params']


        assert inrepr(
            'change.DriverPhueChange',
            driver)

        assert isinstance(
            hash(driver), int)

        assert instr(
            'change.DriverPhueChange',
            driver)


        driver.validate()

        assert driver.plugin

        assert driver.family == family

        assert driver.kinds == ['where']

        assert driver.params


        time = times.current

        assert driver.where(time)



def test_DriverPhueChange_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
    times: 'TestTimes',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    :param times: Times that are used with tests in project.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        devices:
         jupiter_nexist:
          origin: jupiter_philips
          label: Not exists
         neptune_nexist:
          origin: neptune_philips
          label: Not exists
        desires:
         pytest:
          groups:
          - ganymede
          - halimede
          wheres:
          - philips_change:
             devices:
             - jupiter_button
             - neptune_button
             since: 900
          - philips_change:
             devices:
             - jupiter_nexist
             - neptune_nexist
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    childs = homie.childs

    desire = (
        childs.desires
        ['pytest'])


    assert homie.refresh()


    time = times.current

    wheres = desire.wheres

    for where in wheres:

        (where.drivers[0]
         .where(time))



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING

from encommon.times import Time
from encommon.times import Timer
from encommon.types import NCFalse

from .child import HomieChild
from .helpers import occurd
from .helpers import whered
from ..models import HomieModels
from ..plugins import HomieOccur
from ..plugins import HomieWhere

if TYPE_CHECKING:
    from .device import HomieDevice
    from .group import HomieGroup
    from ..params import HomieAspireParams
    from ..threads import HomieStreamItem



class HomieAspire(HomieChild):
    """
    Match specific conditions for determining desired state.
    """

    __occurs: list[HomieOccur]
    __wheres: list[HomieWhere]

    __timer: Timer


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__occurs = []
        self.__wheres = []

        self.__build_occurs()
        self.__build_wheres()


        params = self.params

        timer = Timer(
            params.pause,
            start='min')

        self.__timer = timer


    def __build_occurs(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        occurs = params.occurs
        homie = self.homie

        if occurs is None:
            return None

        model = HomieOccur


        plugins: list[HomieOccur] = []


        for occur in occurs:

            object = model(
                homie, occur)

            plugins.append(object)


        self.__occurs = plugins


    def __build_wheres(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        wheres = params.wheres
        homie = self.homie

        if wheres is None:
            return None

        model = HomieWhere


        plugins: list[HomieWhere] = []


        for where in wheres:

            object = model(
                homie, where)

            plugins.append(object)


        self.__wheres = plugins


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        occurs = self.__occurs

        for occur in occurs:
            occur.validate()

        wheres = self.__wheres

        for where in wheres:
            where.validate()


    @property
    def kind(
        self,
    ) -> Literal['aspire']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'aspire'


    @property
    def params(
        self,
    ) -> 'HomieAspireParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            HomieModels
            .aspire())

        params = super().params

        assert isinstance(
            params, model)

        return params


    @property
    def devices(
        self,
    ) -> list['HomieDevice']:
        """
        Return the value for the attribute from class instance.

        .. note::
           This method is identical to same within HomieDesire.

        :returns: Value for the attribute from class instance.
        """

        params = self.params
        names = params.devices

        devices = (
            self.homie.childs
            .devices)

        childs: set['HomieDevice'] = set()

        if names is None:
            return list(childs)

        for name in names:

            child = devices[name]

            childs.add(child)

        return list(childs)


    @property
    def groups(
        self,
    ) -> list['HomieGroup']:
        """
        Return the value for the attribute from class instance.

        .. note::
           This method is identical to same within HomieDesire.

        :returns: Value for the attribute from class instance.
        """

        params = self.params
        names = params.groups

        groups = (
            self.homie.childs
            .groups)

        childs: set['HomieGroup'] = set()

        if names is None:
            return list(childs)

        for name in names:

            child = groups[name]

            childs.add(child)

        return list(childs)


    @property
    def occurs(
        self,
    ) -> list[HomieOccur]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return list(self.__occurs)


    @property
    def wheres(
        self,
    ) -> list[HomieWhere]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return list(self.__wheres)


    def when(  # noqa: CFQ004
        self,
        sitem: 'HomieStreamItem',
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        .. note::
           Somewhat similar to same method within HomieDesire.

        :param sitem: Item containing information for operation.
        :returns: Boolean indicating the conditional outcomes.
        """

        homie = self.homie

        time = Time('now')


        def _whered() -> bool:

            wheres = (
                whered(self, time))

            if len(wheres) >= 1:
                return all(wheres)

            return True


        def _occurd() -> bool:

            occurs = (
                occurd(self, sitem))

            if len(occurs) >= 1:
                return any(occurs)

            return False


        try:

            wheres = _whered()
            occurs = _occurd()

            return wheres and occurs

        except Exception as reason:

            homie.logger.log_e(
                base=self,
                name=self,
                status='exception',
                exc_info=reason)

        return NCFalse


    @property
    def paused(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether operation delayed.

        .. note::
           Somewhat similar to same method within HomieDesire.

        :returns: Boolean indicating whether operation delayed.
        """

        timer = self.__timer

        return timer.pause(False)


    def matched(
        self,
    ) -> None:
        """
        Perform operations when this desire is the most desired.

        .. note::
           Somewhat similar to same method within HomieDesire.
        """

        homie = self.homie
        timer = self.__timer

        if homie.dryrun:
            return None

        timer.update('now')



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from .addons import HomieAspiredItem
    from .addons import HomieDesiredItem
    from .addons import HomieQueueItem
    from .common import HomieStage
    from .params import HomieParams
    from .params import HomiePrinterParams
    from .params import HomieChildParams
    from .params import HomieOriginParams
    from .params import HomieDeviceParams
    from .params import HomieGroupParams
    from .params import HomieSceneParams
    from .params import HomieAspireParams
    from .params import HomieDesireParams
    from .params import HomiePluginParams
    from .params import HomieOccurParams
    from .params import HomieWhereParams
    from .params import HomieServiceParams
    from .threads import HomieActionItem
    from .threads import HomieStreamItem
    from .threads import HomieThreadItem
    from .threads import HomieUpdateItem
    from ..builtins import BltnModels
    from ..hubitat import HubiModels
    from ..philips import PhueModels
    from ..ubiquiti import UbiqModels



class HomieModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def homie(
        cls,
    ) -> Type['HomieParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieParams)

        return HomieParams


    @classmethod
    def printer(
        cls,
    ) -> Type['HomiePrinterParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomiePrinterParams)

        return HomiePrinterParams


    @classmethod
    def service(
        cls,
    ) -> Type['HomieServiceParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieServiceParams)

        return HomieServiceParams


    @classmethod
    def builtins(
        cls,
    ) -> Type['BltnModels']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..builtins import (
            BltnModels)

        return BltnModels


    @classmethod
    def hubitat(
        cls,
    ) -> Type['HubiModels']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..hubitat import (
            HubiModels)

        return HubiModels


    @classmethod
    def philips(
        cls,
    ) -> Type['PhueModels']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..philips import (
            PhueModels)

        return PhueModels


    @classmethod
    def ubiquiti(
        cls,
    ) -> Type['UbiqModels']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..ubiquiti import (
            UbiqModels)

        return UbiqModels


    @classmethod
    def child(
        cls,
    ) -> Type['HomieChildParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieChildParams)

        return HomieChildParams


    @classmethod
    def origin(
        cls,
    ) -> Type['HomieOriginParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieOriginParams)

        return HomieOriginParams


    @classmethod
    def device(
        cls,
    ) -> Type['HomieDeviceParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieDeviceParams)

        return HomieDeviceParams


    @classmethod
    def group(
        cls,
    ) -> Type['HomieGroupParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieGroupParams)

        return HomieGroupParams


    @classmethod
    def scene(
        cls,
    ) -> Type['HomieSceneParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieSceneParams)

        return HomieSceneParams


    @classmethod
    def desire(
        cls,
    ) -> Type['HomieDesireParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieDesireParams)

        return HomieDesireParams


    @classmethod
    def aspire(
        cls,
    ) -> Type['HomieAspireParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieAspireParams)

        return HomieAspireParams


    @classmethod
    def plugin(
        cls,
    ) -> Type['HomiePluginParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomiePluginParams)

        return HomiePluginParams


    @classmethod
    def occur(
        cls,
    ) -> Type['HomieOccurParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieOccurParams)

        return HomieOccurParams


    @classmethod
    def where(
        cls,
    ) -> Type['HomieWhereParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HomieWhereParams)

        return HomieWhereParams


    @classmethod
    def stage(
        cls,
    ) -> Type['HomieStage']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .common import (
            HomieStage)

        return HomieStage


    @classmethod
    def queue(
        cls,
    ) -> Type['HomieQueueItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .addons import (
            HomieQueueItem)

        return HomieQueueItem


    @classmethod
    def thread(
        cls,
    ) -> Type['HomieThreadItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .threads import (
            HomieThreadItem)

        return HomieThreadItem


    @classmethod
    def action(
        cls,
    ) -> Type['HomieActionItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .threads import (
            HomieActionItem)

        return HomieActionItem


    @classmethod
    def update(
        cls,
    ) -> Type['HomieUpdateItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .threads import (
            HomieUpdateItem)

        return HomieUpdateItem


    @classmethod
    def stream(
        cls,
    ) -> Type['HomieStreamItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .threads import (
            HomieStreamItem)

        return HomieStreamItem


    @classmethod
    def desired(
        cls,
    ) -> Type['HomieDesiredItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .addons import (
            HomieDesiredItem)

        return HomieDesiredItem


    @classmethod
    def aspired(
        cls,
    ) -> Type['HomieAspiredItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .addons import (
            HomieAspiredItem)

        return HomieAspiredItem



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Event
from time import sleep as block_sleep
from typing import Any
from typing import TYPE_CHECKING

from encommon.times import Timer

from .members import HomieActions
from .members import HomieStreams
from .members import HomieUpdates

if TYPE_CHECKING:
    from .homie import Homie
    from .params import HomieServiceParams



_CONGEST = list[tuple[str, int]]



class HomieService:
    """
    Multi-threaded service for processing the desired state.

    :param homie: Primary class instance for Homie Automate.
    """

    __homie: 'Homie'

    __actions: HomieActions
    __updates: HomieUpdates
    __streams: HomieStreams

    __timer: Timer
    __vacate: Event
    __cancel: Event
    __started: bool


    def __init__(
        self,
        homie: 'Homie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.logger.log_d(
            base=self,
            status='initial')

        self.__homie = homie

        self.__build_objects()

        self.__started = False

        homie.logger.log_i(
            base=self,
            status='created')


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        params = self.params

        respite = params.respite


        self.__actions = (
            HomieActions(homie))

        self.__updates = (
            HomieUpdates(homie))

        self.__streams = (
            HomieStreams(homie))


        self.__timer = Timer(
            respite.health,
            start='min')

        self.__vacate = Event()
        self.__cancel = Event()


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'HomieServiceParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        homie = self.__homie
        params = homie.params

        return params.service


    @property
    def actions(
        self,
    ) -> HomieActions:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__actions


    @property
    def updates(
        self,
    ) -> HomieUpdates:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__updates


    @property
    def streams(
        self,
    ) -> HomieStreams:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__streams


    @property
    def running(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined running.

        :returns: List of threads which are determined running.
        """

        running: list[str] = []

        actions = self.__actions
        updates = self.__updates
        streams = self.__streams

        running.extend(
            actions.running)

        running.extend(
            updates.running)

        running.extend(
            streams.running)

        return sorted(running)


    @property
    def zombies(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined zombies.

        :returns: List of threads which are determined zombies.
        """

        zombies: list[str] = []

        actions = self.__actions
        updates = self.__updates
        streams = self.__streams

        zombies.extend(
            actions.zombies)

        zombies.extend(
            updates.zombies)

        zombies.extend(
            streams.zombies)

        return sorted(zombies)


    @property
    def congest(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        congest: _CONGEST = []

        actions = self.__actions
        updates = self.__updates
        streams = self.__streams

        congest.extend(
            actions.congest)

        congest.extend(
            updates.congest)

        congest.extend(
            streams.congest)

        return sorted(congest)


    @property
    def enqueue(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        enqueue: _CONGEST = []

        actions = self.__actions
        updates = self.__updates
        streams = self.__streams

        enqueue.extend(
            actions.enqueue)

        enqueue.extend(
            updates.enqueue)

        enqueue.extend(
            streams.enqueue)

        return sorted(enqueue)


    def start(
        self,
    ) -> None:
        """
        Start the various threads within the Homie class object.
        """

        homie = self.__homie
        started = self.__started

        if started is True:
            return None

        self.__started = True

        homie.logger.log_i(
            base=self,
            status='starting')

        assert homie.refresh()

        self.__streams.start()
        self.__updates.start()
        self.__actions.start()

        homie.logger.log_i(
            base=self,
            status='started')


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        homie = self.__homie

        actions = self.__actions
        updates = self.__updates
        streams = self.__streams

        while self.running:

            updates.operate()

            self.operate_updates()
            self.operate_streams()

            actions.operate()
            streams.operate()

            block_sleep(0.05)

            self.operate_healths()

        homie.logger.log_i(
            base=self,
            status='vacated')


    def operate_updates(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        homie = self.__homie
        childs = homie.childs

        vacate = self.__vacate
        actions = self.__actions
        updates = self.__updates

        uqueue = updates.uqueue

        auqueue = actions.uqueue

        origins = (
            childs.origins
            .values())


        def _set_update() -> None:

            name = uitem.origin
            _name = origin.name

            if name != _name:
                return None

            (origin
             .set_update(uitem))


        while not uqueue.empty:

            uitem = uqueue.get()

            if vacate.is_set():
                continue

            auqueue.put(uitem)

            for origin in origins:
                _set_update()

            homie.printer(uitem)


    def operate_streams(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        homie = self.__homie

        vacate = self.__vacate
        actions = self.__actions
        updates = self.__updates
        streams = self.__streams

        squeue = streams.squeue

        asqueue = actions.squeue
        usqueue = updates.squeue


        while not squeue.empty:

            sitem = squeue.get()

            if vacate.is_set():
                continue

            asqueue.put(sitem)
            usqueue.put(sitem)

            homie.printer(sitem)


    def operate_healths(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        timer = self.__timer

        if timer.pause():
            return None

        vacate = self.__vacate
        cancel = self.__cancel

        self.check_congest()

        if vacate.is_set():
            return None

        if cancel.is_set():
            return None

        if self.check_zombies():
            self.stop()


    def check_zombies(
        self,
    ) -> bool:
        """
        Return the boolean indicating while threads are zombies.

        :returns: Boolean indicating while threads are zombies.
        """

        homie = self.__homie
        zombies = self.zombies

        for name in zombies:

            homie.logger.log_c(
                thread=name,
                status='zombie')

        return len(zombies) != 0


    def check_congest(
        self,
    ) -> bool:
        """
        Return the boolean indicating when queues are congested.

        :returns: Boolean indicating when queues are congested.
        """

        homie = self.__homie
        congest = self.congest

        for name, count in congest:

            homie.logger.log_e(
                queue=name,
                status='congest',
                count=count)

        return len(congest) != 0


    def soft(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Stop the various threads within the Homie class object.

        :param kwargs: Keyword arguments ignored by the method.
        :param args: Positional arguments ignored by the method.
        """

        homie = self.__homie
        started = self.__started
        vacate = self.__vacate

        if started is False:
            return None

        if vacate.is_set():
            return self.stop()

        homie.logger.log_i(
            base=self,
            status='vacating')

        vacate.set()

        self.__streams.stop()
        self.__updates.soft()
        self.__actions.soft()


    def stop(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Stop the various threads within the Homie class object.

        :param kwargs: Keyword arguments ignored by the method.
        :param args: Positional arguments ignored by the method.
        """

        homie = self.__homie
        started = self.__started
        cancel = self.__cancel

        if started is False:
            return None

        if cancel.is_set():
            return None

        cancel.set()

        homie.logger.log_i(
            base=self,
            status='stopping')

        self.__streams.stop()
        self.__updates.stop()
        self.__actions.stop()

        homie.logger.log_i(
            base=self,
            status='stopped')



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from queue import Queue
from typing import Generic
from typing import TYPE_CHECKING
from typing import TypeVar

from encommon.times import Time

if TYPE_CHECKING:
    from ..homie import Homie



HomieQueueItemType = TypeVar('HomieQueueItemType')



@dataclass
class HomieQueueItem:
    """
    Contain information for sharing using the Python queue.
    """

    time: Time


    def __init__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.time = Time()



class HomieQueue(Generic[HomieQueueItemType]):
    """
    Queue object that will allow to handle full conditions.

    :param homie: Primary class instance for Homie Automate.
    :param size: Maximum size for the created queue object.
    """

    __homie: 'Homie'

    __queue: Queue[HomieQueueItemType]


    def __init__(
        self,
        homie: 'Homie',
        size: int = 10000,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie

        self.__queue = Queue(size)


    @property
    def empty(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether the queue is full.

        :returns: Boolean indicating whether the queue is full.
        """

        queue = self.__queue

        return queue.empty()


    @property
    def qsize(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        queue = self.__queue

        return queue.qsize()


    def put(
        self,
        item: HomieQueueItemType,
    ) -> None:
        """
        Store the provided item within the queue class instance.

        :param item: Item containing information for operation.
        """

        homie = self.__homie
        queue = self.__queue

        try:
            queue.put(item, False)

        except Exception as reason:

            homie.logger.log_e(
                base=self,
                status='exception',
                exc_info=reason)


    def get(
        self,
    ) -> HomieQueueItemType:
        """
        Return the next item within the queue in blocking mode.

        :returns: Next item within the queue in blocking mode.
        """

        queue = self.__queue

        return queue.get()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from typing import TYPE_CHECKING

from _pytest.logging import LogCaptureFixture

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...homie import Homie



def test_HomieLogger(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    logger = homie.logger


    attrs = lattrs(logger)

    assert attrs == [
        '_HomieLogger__homie']


    assert inrepr(
        'logger.HomieLogger',
        logger)

    assert isinstance(
        hash(logger), int)

    assert instr(
        'logger.HomieLogger',
        logger)



def test_HomieLogger_message(
    homie: 'Homie',
    caplog: LogCaptureFixture,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param caplog: pytest object for capturing log message.
    """

    logger = homie.logger


    logger.start()

    logger.log_d(about='pytest')
    logger.log_c(about='pytest')
    logger.log_e(about='pytest')
    logger.log_i(about='pytest')
    logger.log_w(about='pytest')

    logger.log(
        level='debug',
        about='pytest')

    logger.stop()

    output = caplog.record_tuples

    assert len(output) == 6

    logger.log_d(about='pytest')
    logger.log_c(about='pytest')
    logger.log_e(about='pytest')
    logger.log_i(about='pytest')
    logger.log_w(about='pytest')

    output = caplog.record_tuples

    assert len(output) == 6



def test_HomieLogger_cover(
    homie: 'Homie',
    caplog: LogCaptureFixture,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param caplog: pytest object for capturing log message.
    """

    logger = homie.logger
    childs = homie.childs
    devices = childs.devices


    thread = Thread(name='foo')

    device = devices[
        'jupiter_light1']


    logger.log_i(base=thread)
    logger.log_i(item=thread)
    logger.log_i(base=homie)
    logger.log_i(item=homie)
    logger.log_i(name=device)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...service import HomieService



def test_HomieUpdates(
    service: 'HomieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    """

    member = service.updates


    attrs = lattrs(member)

    assert attrs == [
        '_HomieMember__homie',
        '_HomieMember__threads',
        '_HomieMember__aqueue',
        '_HomieMember__uqueue',
        '_HomieMember__squeue',
        '_HomieMember__vacate',
        '_HomieMember__cancel']


    assert inrepr(
        'updates.HomieUpdates',
        member)

    assert isinstance(
        hash(member), int)

    assert instr(
        'updates.HomieUpdates',
        member)


    assert member.homie

    assert len(member.threads) == 6

    assert member.aqueue

    assert member.uqueue

    assert member.squeue

    assert member.vacate

    assert member.cancel

    assert len(member.running) == 0


    member.operate()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from pydantic import Field

from .child import HomieChildParams
from .stage import HomieStageParams
from .store import HomieStoreParams
from .where import HomieWhereParams



class HomieDesireParams(HomieChildParams, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    devices: Annotated[
        Optional[list[str]],
        Field(None,
              description='Devices that are in scope',
              min_length=1)]

    groups: Annotated[
        Optional[list[str]],
        Field(None,
              description='Groups that are in scope',
              min_length=1)]

    stage: Annotated[
        Optional[HomieStageParams],
        Field(None,
              description='Default device scene config')]

    scene: Annotated[
        Optional[str],
        Field(None,
              description='Update the group light scene',
              min_length=1)]

    store: Annotated[
        Optional[list[HomieStoreParams]],
        Field(None,
              description='Update values in database',
              min_length=1)]

    wheres: Annotated[
        Optional[list[HomieWhereParams]],
        Field(None,
              description='Plugins for the conditions',
              min_length=1)]

    pause: Annotated[
        int,
        Field(0,
              description='Delay before acting again',
              ge=0, le=86400)]

    delay: Annotated[
        bool,
        Field(True,
              description='Enforce pause on initial')]

    weight: Annotated[
        int,
        Field(0,
              description='Precedence when multiple',
              ge=0, le=1e3)]


    def __init__(
        # NOCVR
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        devices = data.get('devices')
        groups = data.get('groups')

        if isinstance(devices, str):
            data['devices'] = [devices]

        if isinstance(groups, str):
            data['groups'] = [groups]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Event
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import clsname

from ..addons import HomieQueue

if TYPE_CHECKING:
    from ..childs import HomieOrigin
    from ..homie import Homie
    from ..threads import HomieActionItem
    from ..threads import HomieStreamItem
    from ..threads import HomieThread
    from ..threads import HomieUpdateItem



HomieThreads = dict[str, 'HomieThread']

_CONGEST = list[tuple[str, int]]



class HomieMember:
    """
    Common methods and routines for Homie Automate members.

    :param homie: Primary class instance for Homie Automate.
    """

    __homie: 'Homie'

    __threads: HomieThreads

    __aqueue: HomieQueue['HomieActionItem']
    __uqueue: HomieQueue['HomieUpdateItem']
    __squeue: HomieQueue['HomieStreamItem']
    __vacate: Event
    __cancel: Event


    def __init__(
        self,
        homie: 'Homie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.logger.log_d(
            base=self,
            status='initial')

        self.__homie = homie

        self.__threads = {}

        self.__build_objects()

        self.__post__()

        homie.logger.log_i(
            base=self,
            status='created')


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie

        self.__aqueue = (
            HomieQueue(homie))

        self.__uqueue = (
            HomieQueue(homie))

        self.__squeue = (
            HomieQueue(homie))

        self.__vacate = Event()
        self.__cancel = Event()

        self.__build_threads()


    def __build_threads(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        childs = homie.childs
        origins = childs.origins


        threads: 'HomieThreads' = {}


        items = origins.items()

        for name, origin in items:

            thread = (
                self
                .get_thread(origin))

            if thread is None:
                continue

            object = thread(
                self, origin)

            threads[name] = object


        self.__threads = threads


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def threads(
        self,
    ) -> HomieThreads:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        threads = self.__threads

        return dict(threads)


    @property
    def aqueue(
        self,
    ) -> HomieQueue['HomieActionItem']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__aqueue


    @property
    def uqueue(
        self,
    ) -> HomieQueue['HomieUpdateItem']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__uqueue


    @property
    def squeue(
        self,
    ) -> HomieQueue['HomieStreamItem']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__squeue


    @property
    def vacate(
        self,
    ) -> Event:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__vacate


    @property
    def cancel(
        self,
    ) -> Event:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__cancel


    @property
    def running(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined running.

        :returns: List of threads which are determined running.
        """

        running: list[str] = []

        threads = (
            self.__threads
            .values())

        for thread in threads:

            if not thread.is_alive():
                continue

            name = thread.name

            running.append(name)

        return sorted(running)


    @property
    def zombies(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined zombies.

        :returns: List of threads which are determined zombies.
        """

        zombies: list[str] = []

        threads = (
            self.__threads
            .values())

        for thread in threads:

            if thread.is_alive():
                continue

            name = thread.name

            zombies.append(name)

        return sorted(zombies)


    @property
    def congest(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        congest: _CONGEST = []

        name = clsname(self)

        aqueue = self.aqueue
        uqueue = self.uqueue
        squeue = self.squeue


        if aqueue.qsize > 5:

            append = (
                f'{name}/aqueue',
                aqueue.qsize)

            congest.append(append)

        if uqueue.qsize > 5:

            append = (
                f'{name}/uqueue',
                uqueue.qsize)

            congest.append(append)

        if squeue.qsize > 5:

            append = (
                f'{name}/squeue',
                squeue.qsize)

            congest.append(append)


        threads = (
            self.__threads
            .values())

        for thread in threads:

            congest.extend(
                thread.congest)


        return sorted(congest)


    @property
    def enqueue(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        enqueue: _CONGEST = []

        name = clsname(self)

        aqueue = self.aqueue
        uqueue = self.uqueue
        squeue = self.squeue


        if aqueue.qsize > 0:

            append = (
                f'{name}/aqueue',
                aqueue.qsize)

            enqueue.append(append)

        if uqueue.qsize > 0:

            append = (
                f'{name}/uqueue',
                uqueue.qsize)

            enqueue.append(append)

        if squeue.qsize > 0:

            append = (
                f'{name}/squeue',
                squeue.qsize)

            enqueue.append(append)


        threads = (
            self.__threads
            .values())

        for thread in threads:

            enqueue.extend(
                thread.enqueue)


        return sorted(enqueue)


    def start(
        self,
    ) -> None:
        """
        Start the various threads within the Homie class object.
        """

        threads = (
            self.__threads
            .values())

        for thread in threads:
            thread.start()


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        raise NotImplementedError


    def operate_updates(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        # Where the updates spread

        vacate = self.vacate
        uqueue = self.uqueue

        threads = (
            self.__threads
            .values())


        def _put_update() -> None:

            uqueue = thread.uqueue

            uqueue.put(uitem)


        while not uqueue.empty:

            uitem = uqueue.get()

            if vacate.is_set():
                continue

            for thread in threads:
                _put_update()


    def soft(
        self,
    ) -> None:
        """
        Stop the various threads within the Homie class object.
        """

        self.__vacate.set()


    def stop(
        self,
    ) -> None:
        """
        Stop the various threads within the Homie class object.
        """

        self.__cancel.set()

        threads = (
            self.__threads
            .values())

        for thread in threads:
            thread.stop()


    def get_thread(
        self,
        origin: 'HomieOrigin',
    ) -> Optional[Type['HomieThread']]:
        """
        Return the Homie class definition for its instantiation.

        :param origin: Child class instance for Homie Automate.
        :returns: Homie class definition for its instantiation.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from .member import HomieMember
from ...philips import PhueStream

if TYPE_CHECKING:
    from ..childs import HomieOrigin
    from ..threads import HomieThread



class HomieStreams(HomieMember):
    """
    Common methods and routines for Homie Automate members.
    """


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        # Nothing to do for member


    def get_thread(
        self,
        origin: 'HomieOrigin',
    ) -> Optional[Type['HomieThread']]:
        """
        Return the Homie class definition for its instantiation.

        :param origin: Child class instance for Homie Automate.
        :returns: Homie class definition for its instantiation.
        """

        family = origin.family

        if family == 'philips':
            return PhueStream

        return None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from .params import DriverBltnPeriodParams
    from .params import DriverBltnRegexpParams
    from .params import DriverBltnStoreParams



class BltnDriverModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def store(
        cls,
    ) -> Type['DriverBltnStoreParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverBltnStoreParams)

        return DriverBltnStoreParams


    @classmethod
    def period(
        cls,
    ) -> Type['DriverBltnPeriodParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverBltnPeriodParams)

        return DriverBltnPeriodParams


    @classmethod
    def regexp(
        cls,
    ) -> Type['DriverBltnRegexpParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverBltnRegexpParams)

        return DriverBltnRegexpParams



class BltnModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def drivers(
        cls,
    ) -> Type['BltnDriverModels']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        return BltnDriverModels



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from argparse import ArgumentParser
from signal import SIGHUP
from signal import SIGINT
from signal import SIGTERM
from signal import signal
from sys import argv as sys_argv
from typing import Optional

from encommon.types import DictStrAny

from ..homie import Homie
from ..homie import HomieConfig
from ..homie import HomieService



def arguments(  # noqa: CFQ001
    args: Optional[list[str]] = None,
) -> DictStrAny:
    """
    Construct arguments which are associated with the file.

    :param args: Override the source for the main arguments.
    :returns: Construct arguments from command line options.
    """

    parser = ArgumentParser()

    args = args or sys_argv[1:]


    parser.add_argument(
        '--config',
        required=True,
        help=(
            'complete or relative '
            'path to config file'))


    parser.add_argument(
        '--console',
        action='store_true',
        default=False,
        help=(
            'write log messages '
            'to standard output'))


    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help=(
            'increase logging level '
            'for standard output'))


    parser.add_argument(
        '--dry-run',
        action='store_true',
        default=False,
        dest='dryrun',
        help='do not execute actions')


    parser.add_argument(
        '--idempotent',
        action='store_false',
        default=None,
        dest='potent',
        help=(
            'do not make requests '
            'when already applied'))


    parser.add_argument(
        '--respite_desire',
        type=int,
        dest='idesire',
        help=(
            'override the interval '
            'for desires schedule'))


    parser.add_argument(
        '--respite_update',
        type=int,
        dest='iupdate',
        help=(
            'override the interval '
            'for updates schedule'))


    parser.add_argument(
        '--timeout_action',
        type=int,
        dest='atimeout',
        help=(
            'override the time '
            'for action request'))


    parser.add_argument(
        '--timeout_update',
        type=int,
        dest='utimeout',
        help=(
            'override the time '
            'for update request'))


    parser.add_argument(
        '--timeout_stream',
        type=int,
        dest='stimeout',
        help=(
            'override the timeout '
            'for stream request'))


    parser.add_argument(
        '--print_action',
        action='store_true',
        default=False,
        dest='paction',
        help=(
            'print the submited '
            'actions to console'))


    parser.add_argument(
        '--print_update',
        action='store_true',
        default=False,
        dest='pupdate',
        help=(
            'print the streamed '
            'updates to console'))


    parser.add_argument(
        '--print_stream',
        action='store_true',
        default=False,
        dest='pstream',
        help=(
            'print the streamed '
            'events to console'))


    parser.add_argument(
        '--print_desire',
        action='store_true',
        default=False,
        dest='pdesire',
        help=(
            'print the desired '
            'state to console'))


    parser.add_argument(
        '--print_aspire',
        action='store_true',
        default=False,
        dest='paspire',
        help=(
            'print the desired '
            'state to console'))


    return vars(
        parser
        .parse_args(args))



def operation(
    # NOCVR
    homie: Homie,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param homie: Primary class instance for Homie Automate.
    """

    service = HomieService(homie)

    service.start()

    signal(SIGINT, service.soft)
    signal(SIGTERM, service.soft)
    signal(SIGHUP, service.soft)

    service.operate()



def execution(
    # NOCVR
    args: Optional[list[str]] = None,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param args: Override the source for the main arguments.
    """

    config = HomieConfig(
        arguments(args))

    config.logger.start()

    config.logger.log_i(
        base='execution/service',
        status='started')

    homie = Homie(config)

    operation(homie)

    config.logger.log_i(
        base='execution/service',
        status='stopped')

    config.logger.stop()



if __name__ == '__main__':
    execution()  # NOCVR



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from typing import Any
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import clsname

from .queue import HomieQueueItem

if TYPE_CHECKING:
    from ..homie import Homie



class HomieLogger:
    """
    Methods for extending use of underlying logging library.
    """

    __homie: 'Homie'


    def __init__(
        self,
        homie: 'Homie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie


    def start(
        self,
    ) -> None:
        """
        Initialize the Python logging library using parameters.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        logger.start()


    def stop(
        self,
    ) -> None:
        """
        Deinitialize the Python logging library using parameters.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        logger.stop()


    def __process(
        self,
        kwargs: DictStrAny,
    ) -> None:
        """
        Process the keyword arguments handling relevant values.

        :param kwargs: Keyword arguments for populating message.
        """

        base = kwargs.get('base')
        item = kwargs.get('item')
        name = kwargs.get('name')

        threads = (
            HomieQueueItem,
            Thread)

        prefix = '<enhomie.'


        if isinstance(base, threads):
            kwargs['base'] = (
                clsname(base))

        elif str(base)[:9] == prefix:
            kwargs['base'] = (
                clsname(base))


        if isinstance(item, threads):
            kwargs['item'] = (
                clsname(item))

        elif str(item)[:9] == prefix:
            kwargs['item'] = (
                clsname(item))


        if (hasattr(name, 'name')
                and name is not None):
            kwargs['name'] = name.name


    def log(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log`.

        :param kwargs: Keyword arguments for populating message.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        self.__process(kwargs)

        logger.log(**kwargs)


    def log_c(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_c`.

        :param kwargs: Keyword arguments for populating message.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        self.__process(kwargs)

        logger.log_c(**kwargs)


    def log_d(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_d`.

        :param kwargs: Keyword arguments for populating message.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        self.__process(kwargs)

        logger.log_d(**kwargs)


    def log_e(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_e`.

        :param kwargs: Keyword arguments for populating message.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        self.__process(kwargs)

        logger.log_e(**kwargs)


    def log_i(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_i`.

        :param kwargs: Keyword arguments for populating message.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        self.__process(kwargs)

        logger.log_i(**kwargs)


    def log_w(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_w`.

        :param kwargs: Keyword arguments for populating message.
        """

        homie = self.__homie
        config = homie.config
        logger = config.logger

        self.__process(kwargs)

        logger.log_w(**kwargs)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .actions import HomieActions
from .member import HomieMember
from .streams import HomieStreams
from .updates import HomieUpdates



__all__ = [
    'HomieMember',
    'HomieActions',
    'HomieUpdates',
    'HomieStreams']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Any
from typing import TYPE_CHECKING
from typing import Union

from encommon.times import Timer
from encommon.types import NCNone

from .thread import HomieThread
from .thread import HomieThreadItem

if TYPE_CHECKING:
    from .stream import HomieStreamItem
    from ..childs import HomieOrigin



HomieUpdateBase = Union[
    'HomieUpdateItem',
    'HomieStreamItem']

_UPDATES = set['HomieOrigin']



@dataclass
class HomieUpdateItem(HomieThreadItem):
    """
    Contain information for sharing using the Python queue.
    """


    def __init__(
        self,
        origin: 'HomieOrigin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(origin)



class HomieUpdate(HomieThread):
    """
    Common methods and routines for Homie Automate threads.
    """

    __timer: Timer


    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(
            *args, **kwargs)


        homie = self.homie
        params = homie.params

        respite = (
            params.service
            .respite)

        timer = Timer(
            respite.update,
            start='min')

        self.__timer = timer


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        member = self.member
        timer = self.__timer
        vacate = member.vacate

        if timer.pause():
            return None

        if vacate.is_set():
            return None

        homie = self.homie
        params = homie.params
        origin = self.origin

        timeout = (
            params.service
            .timeout.update)

        refresh = (
            origin
            .refresh(timeout))

        if refresh is False:
            return NCNone

        origin.put_update(
            member.uqueue)


    def operate_updates(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        member = self.member
        homie = member.homie
        childs = homie.childs
        vacate = member.vacate
        uqueue = member.uqueue
        squeue = self.squeue

        origins = (
            childs.origins
            .values())


        updates: _UPDATES = set()


        def _set_update() -> None:

            name = sitem.origin
            _name = origin.name

            if name != _name:
                return None

            result = (
                origin
                .set_update(sitem))

            if result is False:
                return NCNone

            updates.add(origin)


        while not squeue.empty:

            sitem = squeue.get()

            if vacate.is_set():
                continue

            if self.expired(sitem):
                continue

            for origin in origins:
                _set_update()


        for origin in updates:

            (origin
             .put_update(uqueue))


        super().operate_updates()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import DictStrAny

from httpx import ReadTimeout

from .origin import PhueOrigin
from ..homie.threads import HomieStream
from ..homie.threads import HomieStreamItem

if TYPE_CHECKING:
    from ..homie.childs import HomieOrigin



@dataclass
class PhueStreamItem(HomieStreamItem):
    """
    Contain information for sharing using the Python queue.
    """

    event: DictStrAny


    def __init__(
        self,
        origin: 'HomieOrigin',
        event: DictStrAny,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(
            origin, event)



class PhueStream(HomieStream):
    """
    Common methods and routines for Homie Automate threads.
    """


    def operate_streams(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        homie = self.homie
        member = self.member
        origin = self.origin

        assert isinstance(
            origin, PhueOrigin)

        params = homie.params
        squeue = member.squeue
        cancel = member.cancel

        request = (
            origin.bridge
            .events_block)

        timeout = (
            params.service
            .timeout.stream)

        stream = request(
            timeout=timeout)

        try:

            homie.logger.log_i(
                base=self,
                name=origin,
                status='reading')

            for event in stream:

                if cancel.is_set():
                    break

                origin.put_stream(
                    squeue, event)

            homie.logger.log_i(
                base=self,
                name=origin,
                status='closed')

        except ReadTimeout:

            homie.logger.log_i(
                base=self,
                name=origin,
                status='timeout')

        except Exception as reason:

            homie.logger.log_e(
                base=self,
                name=origin,
                status='exception',
                exc_info=reason)

            block_sleep(1)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import save_text

from respx import MockRouter

from ....conftest import config_factory
from ....conftest import homie_factory

if TYPE_CHECKING:
    from ...homie import Homie
    from ....utils import TestBodies



def test_HomieGroup(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    groups = childs.groups


    moons = (
        dict(bodies.moons)
        .items())

    for planet, moon in moons:

        origin = origins[
            f'{planet}_philips']

        group = groups[moon]


        attrs = lattrs(group)

        assert attrs == [
            '_HomieChild__homie',
            '_HomieChild__name',
            '_HomieChild__params']


        assert inrepr(
            'group.HomieGroup',
            group)

        assert isinstance(
            hash(group), int)

        assert instr(
            'group.HomieGroup',
            group)


        group.validate()

        assert group.homie

        assert group.name == moon

        assert group.family == 'builtins'

        assert group.kind == 'group'

        assert group.origin

        assert group.params

        assert not group.source

        assert group.dumped


        assert origin.refresh()

        assert group.source



def test_HomieGroup_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        groups:
         pytest:
          devices:
          - jupiter_special
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    childs = homie.childs

    group = (
        childs.groups
        ['pytest'])

    assert not group.source



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import NCFalse
from encommon.types import merge_dicts
from encommon.types import strplwr

if TYPE_CHECKING:
    from ..homie.common import HomieKinds



UbiqFetch = DictStrAny
UbiqMerge = dict[str, UbiqFetch]

UbiqKinds = Literal[
    'historic',
    'realtime']



def merge_fetch(
    fetch: UbiqFetch,
) -> UbiqMerge:
    """
    Return the content related to the item in parent system.

    :param fetch: Dictionary from the origin to be updated.
    :returns: Content related to the item in parent system.
    """

    fetch = deepcopy(fetch)


    staged: UbiqMerge = {}
    source: UbiqMerge = {}

    historic = (
        fetch['historic']
        ['data'])

    realtime = (
        fetch['realtime']
        ['data'])


    def _fixtimes(
        fetch: UbiqFetch,
    ) -> None:

        if 'first_seen' in item:
            item['first_seen'] = [
                item['first_seen']]

        if 'last_seen' in item:
            item['last_seen'] = [
                item['last_seen']]


    def _combine(
        target: UbiqKinds,
    ) -> None:

        unique = item['_id']

        if unique not in staged:
            staged[unique] = {}

        source = staged[unique]

        assert target not in source

        _fixtimes(item)

        source[target] = (
            deepcopy(item))


    for item in historic:
        _combine('historic')

    for item in realtime:
        _combine('realtime')


    items = staged.items()

    for unique, fetch in items:

        historic = deepcopy(
            fetch.get('historic')
            or {})

        realtime = deepcopy(
            fetch.get('realtime')
            or {})

        merge_dicts(
            dict1=historic,
            dict2=realtime,
            force=True)

        value = historic | {
            '_source': fetch}

        source[unique] = value


    _source = sorted(
        source.items())

    return dict(_source)



def merge_find(  # noqa: CFQ004
    merge: UbiqMerge,
    kind: Optional['HomieKinds'] = None,
    unique: Optional[str] = None,
    label: Optional[str] = None,
) -> list[DictStrAny]:
    """
    Return the content related to the item in parent system.

    :param merge: Dictionary from the origin to be updated.
    :param kind: Which kind of Homie object will be located.
    :param unique: Unique identifier within parents system.
    :param label: Friendly name or label within the parent.
    :returns: Content related to the item in parent system.
    """

    assert unique or label

    if label is not None:
        label = strplwr(label)

    if unique is not None:
        unique = strplwr(unique)

    found: list[DictStrAny] = []


    if kind is not None:
        assert kind == 'device'


    def _match_unique() -> bool:

        if unique is None:
            return True

        uniques = [
            value.get('_id'),
            value.get('last_ip'),
            value.get('mac')]

        uniques = [
            strplwr(x)
            for x in uniques
            if x is not None]

        return unique in uniques


    def _match_label() -> bool:

        if label is None:
            return True

        if 'name' not in value:
            return NCFalse

        _label = strplwr(
            value['name'])

        return label == _label


    values = merge.values()

    for value in values:

        if not _match_unique():
            continue

        if not _match_label():
            continue

        found.append(value)


    return found
